from __future__ import annotations

from tach.parsing.config import (
    dump_project_config_to_yaml,
    parse_package_config,
    parse_project_config,
)
from tach.parsing.imports import get_project_imports
from tach.parsing.interface import parse_interface_members
from tach.parsing.packages import build_package_trie

__all__ = [
    "parse_package_config",
    "parse_project_config",
    "dump_project_config_to_yaml",
    "get_project_imports",
    "parse_interface_members",
    "build_package_trie",
]



from __future__ import annotations

import uuid
from pathlib import Path
from typing import Optional

from tach.cache.setup import resolve_dot_tach
from tach.filesystem import find_project_config_root


def get_uid() -> Optional[uuid.UUID]:
    project_root = find_project_config_root(str(Path.cwd()))
    if project_root is None:
        return
    project_path = Path(project_root)
    if not (project_path / ".tach" / "tach.info").exists():
        resolve_dot_tach()
    with open(project_path / ".tach" / "tach.info") as f:
        uid = uuid.UUID(f.read().strip())
    return uid



from __future__ import annotations

import uuid
from pathlib import Path
from typing import Optional

from tach.filesystem import find_project_config_root


def resolve_dot_tach() -> Optional[Path]:
    project_dir = find_project_config_root(str(Path.cwd()))
    if project_dir is None:
        return
    project_path = Path(project_dir)

    def _create(path: Path, is_file: bool = False, file_content: str = "") -> None:
        if not path.exists():
            if is_file:
                path.write_text(file_content.strip())
            else:
                path.mkdir()

    # Create .tach
    tach_path = project_path / ".tach"
    _create(tach_path)
    # Create info
    info_path = tach_path / "tach.info"
    _create(info_path, is_file=True, file_content=str(uuid.uuid4()))
    # Create .gitignore
    gitignore_content = """
# This folder is for tach. Do not edit.

# gitignore all content, including this .gitignore
*
    """
    gitignore_path = tach_path / ".gitignore"
    _create(gitignore_path, is_file=True, file_content=gitignore_content)
    return Path(tach_path)



from __future__ import annotations

PACKAGE_NAME = "tach"
TOOL_NAME = "tach"
CONFIG_FILE_NAME = TOOL_NAME
PACKAGE_FILE_NAME = "package"

__all__ = ["PACKAGE_NAME", "TOOL_NAME", "CONFIG_FILE_NAME", "PACKAGE_FILE_NAME"]



from __future__ import annotations

from typing import Any, List, Optional

from pydantic import BaseModel, Field


class Config(BaseModel):
    model_config = {"extra": "forbid"}


class PackageConfig(Config):
    """
    Configuration for a single package within a project.
    """

    tags: List[str]
    strict: bool = False


class TagDependencyRules(Config):
    """
    Dependency rules for a particular set of tags (typically one tag).
    """

    tag: str
    depends_on: List[str]


def is_deprecated_project_config(config: dict[str, Any]) -> bool:
    if not config:
        return False
    if "exclude_hidden_paths" in config:
        return True
    if "constraints" in config and not (
        set(config.keys()) - {"constraints", "exclude"}
    ):
        # This appears to be a project config object,
        # the deprecated version will have a dict of constraints
        return isinstance(config["constraints"], dict)
    return False


def fix_deprecated_config(config: dict[str, Any]):
    if "constraints" in config and isinstance(config["constraints"], dict):
        config["constraints"] = [
            {"tag": key, **value}
            for key, value in config.get("constraints", {}).items()
        ]
    if "exclude_hidden_paths" in config:
        config.pop("exclude_hidden_paths")


class ProjectConfig(Config):
    """
    Configuration applied globally to a project.
    """

    constraints: List[TagDependencyRules] = Field(default_factory=list)
    exclude: Optional[List[str]] = Field(default_factory=lambda: ["tests", "docs"])
    exact: bool = False
    disable_logging: bool = False
    ignore_type_checking_imports: bool = False

    def dependencies_for_tag(self, tag: str) -> list[str]:
        return next(
            (
                constraint.depends_on
                for constraint in self.constraints
                if constraint.tag == tag
            ),
            [],  # type: ignore
        )

    def add_dependencies_to_tags(self, tags: list[str], dependencies: list[str]):
        for tag in tags:
            current_dependency_rules = next(
                (
                    constraint
                    for constraint in self.constraints
                    if constraint.tag == tag
                ),
                None,
            )
            if not current_dependency_rules:
                # No constraint exists for tag, just add the new dependencies
                self.constraints.append(
                    TagDependencyRules(tag=tag, depends_on=dependencies)
                )
            else:
                # Constraints already exist, set the union of existing and new as dependencies
                new_dependencies = set(current_dependency_rules.depends_on) | set(
                    dependencies
                )
                current_dependency_rules.depends_on = list(new_dependencies)

    def find_extra_constraints(
        self, other_config: "ProjectConfig"
    ) -> list[TagDependencyRules]:
        extra_constraints: list[TagDependencyRules] = []
        base_constraint_tags = set(constraint.tag for constraint in self.constraints)
        for constraint in other_config.constraints:
            if constraint.tag not in base_constraint_tags:
                extra_constraints.append(constraint)
                continue
            base_constraint_dependencies = set(
                self.dependencies_for_tag(constraint.tag)
            )
            extra_dependencies = (
                set(other_config.dependencies_for_tag(constraint.tag))
                - base_constraint_dependencies
            )
            if extra_dependencies:
                extra_constraints.append(
                    TagDependencyRules(
                        tag=constraint.tag,
                        depends_on=list(extra_dependencies),
                    )
                )

        return extra_constraints

    @classmethod
    def factory(cls, config: dict[str, Any]) -> tuple[bool, "ProjectConfig"]:
        """
        Using this factory to catch deprecated config and flag it to the caller
        """
        if is_deprecated_project_config(config):
            fix_deprecated_config(config)
            return True, ProjectConfig(**config)
        return False, ProjectConfig(**config)  # type: ignore



from __future__ import annotations

import os
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional

from tach import errors
from tach import filesystem as fs
from tach.parsing import build_package_trie, get_project_imports

if TYPE_CHECKING:
    from tach.core import PackageNode, PackageTrie, ProjectConfig


@dataclass
class ErrorInfo:
    source_tags: list[str] = field(default_factory=list)
    invalid_tags: list[str] = field(default_factory=list)
    allowed_tags: list[str] = field(default_factory=list)
    exception_message: str = ""

    @property
    def is_tag_error(self) -> bool:
        return all((self.source_tags, self.invalid_tags))


def is_top_level_package_import(mod_path: str, package: PackageNode) -> bool:
    return mod_path == package.full_path


def import_matches_interface_members(mod_path: str, package: PackageNode) -> bool:
    mod_path_segments = mod_path.rsplit(".", 1)
    if len(mod_path_segments) == 1:
        return mod_path_segments[0] == package.full_path
    else:
        mod_pkg_path, mod_member_name = mod_path_segments
        return (
            mod_pkg_path == package.full_path
            and mod_member_name in package.interface_members
        )


def check_import(
    project_config: ProjectConfig,
    package_trie: PackageTrie,
    import_mod_path: str,
    file_mod_path: str,
    file_nearest_package: Optional[PackageNode] = None,
) -> Optional[ErrorInfo]:
    import_nearest_package = package_trie.find_nearest(import_mod_path)
    if import_nearest_package is None:
        # This shouldn't happen since we intend to filter out any external imports,
        # but we should allow external imports if they have made it here.
        return None

    # Lookup file_mod_path if package not given
    if file_nearest_package is None:
        file_nearest_package = package_trie.find_nearest(file_mod_path)
    # If package not found, we should fail since the implication is that
    # an external package is importing directly from our project
    if file_nearest_package is None:
        return ErrorInfo(
            exception_message=f"Package containing '{file_mod_path}' not found in project.",
        )

    # Imports within the same package are always allowed
    if import_nearest_package == file_nearest_package:
        return None

    import_package_config = import_nearest_package.config
    if import_package_config and import_package_config.strict:
        if not is_top_level_package_import(
            import_mod_path, import_nearest_package
        ) and not import_matches_interface_members(
            import_mod_path, import_nearest_package
        ):
            # In strict mode, import must be of the package itself or one of the
            # interface members (defined in __all__)
            return ErrorInfo(
                exception_message=(
                    f"Package '{import_nearest_package.full_path}' is in strict mode. "
                    "Only imports from the root of this package are allowed. "
                    f"The import '{import_mod_path}' (in '{file_mod_path}') "
                    f"is not included in __all__."
                ),
            )

    # The import must be explicitly allowed based on the tags and top-level config
    if not file_nearest_package.config or not import_nearest_package.config:
        return ErrorInfo(
            exception_message="Could not find package configuration.",
        )
    file_tags = file_nearest_package.config.tags
    import_tags = import_nearest_package.config.tags

    for file_tag in file_tags:
        dependency_tags = project_config.dependencies_for_tag(file_tag)
        if any(
            any(dependency_tag == import_tag for dependency_tag in dependency_tags)
            for import_tag in import_tags
        ):
            # The import has at least one tag which matches at least one expected dependency
            continue
        # This means the import has no tags which the file can depend on
        return ErrorInfo(
            source_tags=file_tags,
            invalid_tags=import_tags,
            allowed_tags=dependency_tags,
        )

    return None


@dataclass
class BoundaryError:
    file_path: str
    line_number: int
    import_mod_path: str
    error_info: ErrorInfo


def check(
    root: str,
    project_config: ProjectConfig,
    exclude_paths: Optional[list[str]] = None,
) -> list[BoundaryError]:
    if not os.path.isdir(root):
        raise errors.TachSetupError(f"The path {root} is not a valid directory.")

    cwd = fs.get_cwd()
    try:
        fs.chdir(root)
        # This 'canonicalizes' the path arguments, resolving directory traversal
        root = fs.canonical(root)

        if exclude_paths is not None and project_config.exclude is not None:
            exclude_paths.extend(project_config.exclude)
        else:
            exclude_paths = project_config.exclude

        package_trie = build_package_trie(
            root,
            exclude_paths=exclude_paths,
        )

        boundary_errors: list[BoundaryError] = []
        for file_path in fs.walk_pyfiles(
            root,
            exclude_paths=exclude_paths,
        ):
            mod_path = fs.file_to_module_path(file_path)
            nearest_package = package_trie.find_nearest(mod_path)
            if nearest_package is None:
                continue

            # This should only give us imports from within our project
            # (excluding stdlib, builtins, and 3rd party packages)
            project_imports = get_project_imports(
                root,
                file_path,
                ignore_type_checking_imports=project_config.ignore_type_checking_imports,
            )
            for project_import in project_imports:
                check_error = check_import(
                    project_config=project_config,
                    package_trie=package_trie,
                    import_mod_path=project_import.mod_path,
                    file_nearest_package=nearest_package,
                    file_mod_path=mod_path,
                )
                if check_error is None:
                    continue

                boundary_errors.append(
                    BoundaryError(
                        file_path=file_path,
                        import_mod_path=project_import.mod_path,
                        line_number=project_import.line_number,
                        error_info=check_error,
                    )
                )

        return boundary_errors
    finally:
        fs.chdir(cwd)



from __future__ import annotations

from pathlib import Path

from tach.filesystem.service import mark_executable, write_file
from tach.hooks import build_pre_commit_hook_content


def install_pre_commit(path: str = ".", project_root: str = "") -> tuple[bool, str]:
    root_path = Path(path)

    git_hooks_dir = root_path / ".git/hooks"
    hook_dst = git_hooks_dir / "pre-commit"

    if not git_hooks_dir.exists():
        return False, f"'{git_hooks_dir}' directory does not exist"

    if hook_dst.exists():
        return False, f"'{hook_dst}' already exists, you'll need to install manually"

    pre_commit_hook_content = build_pre_commit_hook_content(root=project_root)

    write_file(str(hook_dst), pre_commit_hook_content)
    mark_executable(str(hook_dst))
    return True, ""



from __future__ import annotations

import os
from typing import Optional

from tach import errors
from tach import filesystem as fs
from tach.check import check
from tach.colors import BCOLORS
from tach.constants import CONFIG_FILE_NAME
from tach.core import ProjectConfig
from tach.parsing import dump_project_config_to_yaml, parse_project_config


def sync_dependency_constraints(
    root: str,
    project_config: ProjectConfig,
    filter_tags: Optional[set[str]] = None,
    exclude_paths: Optional[list[str]] = None,
) -> ProjectConfig:
    """
    Update project configuration with auto-detected dependency constraints.
    This is additive, meaning it will create dependencies to resolve existing errors,
    but will not remove any constraints.
    Passing 'filter_tags' will limit updates to only those dependencies which include one of those tags.
    """
    check_errors = check(
        root, project_config=project_config, exclude_paths=exclude_paths
    )
    for error in check_errors:
        error_info = error.error_info
        if error_info.is_tag_error:
            if not filter_tags:
                project_config.add_dependencies_to_tags(
                    error_info.source_tags, error_info.invalid_tags
                )
            else:
                source_tags = set(error_info.source_tags)
                invalid_tags = set(error_info.invalid_tags)
                if source_tags & filter_tags:
                    # A package with one of the added tags caused this error and should update its dependencies
                    project_config.add_dependencies_to_tags(
                        error_info.source_tags, error_info.invalid_tags
                    )
                if invalid_tags & filter_tags:
                    # A package now depends on one of the added tags and should add the newly added tags
                    # Note that we should leave pre-existing invalid tags
                    project_config.add_dependencies_to_tags(
                        error_info.source_tags, list(invalid_tags & filter_tags)
                    )

    return project_config


def prune_dependency_constraints(
    root: str,
    project_config: Optional[ProjectConfig] = None,
    exclude_paths: Optional[list[str]] = None,
) -> ProjectConfig:
    """
    Build a minimal project configuration with auto-detected dependency constraints.
    """
    if project_config is not None:
        # Force constraints to be empty in case we received configuration with pre-existing constraints
        project_config = project_config.model_copy(update={"constraints": []})
    else:
        project_config = ProjectConfig()

    sync_dependency_constraints(
        root, project_config=project_config, exclude_paths=exclude_paths
    )

    return project_config


def sync_project(
    prune: bool = False, exclude_paths: Optional[list[str]] = None
) -> None:
    original_cwd = fs.get_cwd()
    try:
        root = fs.find_project_config_root(original_cwd)
        if root is None:
            raise errors.TachSetupError(
                f"{BCOLORS.WARNING}Could not find project configuration.{BCOLORS.ENDC}"
            )
        fs.chdir(root)
        project_config = parse_project_config(root=root)
        if exclude_paths is not None and project_config.exclude is not None:
            exclude_paths.extend(project_config.exclude)
        else:
            exclude_paths = project_config.exclude

        if prune:
            project_config = prune_dependency_constraints(
                root, project_config=project_config, exclude_paths=exclude_paths
            )
        else:
            project_config = sync_dependency_constraints(
                root, project_config=project_config, exclude_paths=exclude_paths
            )

        tach_yml_path = os.path.join(root, f"{CONFIG_FILE_NAME}.yml")
        tach_yml_content = dump_project_config_to_yaml(project_config)
        fs.write_file(tach_yml_path, tach_yml_content)

    finally:
        fs.chdir(original_cwd)



from __future__ import annotations

import os

import pytest
from pydantic import ValidationError

from tach import filesystem as fs
from tach.check import check
from tach.core.config import PackageConfig, ProjectConfig, TagDependencyRules
from tach.filesystem import file_to_module_path
from tach.parsing.config import parse_package_config, parse_project_config


def test_file_to_mod_path():
    assert file_to_module_path("__init__.py") == ""
    assert file_to_module_path("domain_one/__init__.py") == "domain_one"
    assert file_to_module_path("domain_one/interface.py") == "domain_one.interface"


def test_parse_valid_project_config():
    result = parse_project_config("example/valid/")
    assert result == ProjectConfig(
        constraints=[
            TagDependencyRules(tag="one", depends_on=["two"]),
            TagDependencyRules(tag="two", depends_on=["one"]),
            TagDependencyRules(tag="three", depends_on=[]),
        ],
        exclude=["domain_thr.*"],
    )


def test_run_valid_project_config():
    current_dir = os.getcwd()
    try:
        project = "./example/valid/"
        fs.chdir(project)
        project_config = parse_project_config()
        results = check(
            ".",
            project_config,
            exclude_paths=project_config.exclude,
        )
        assert results == []
    finally:
        # Make sure not to dirty the test directory state
        fs.chdir(current_dir)


def test_parse_valid_strict_package_config():
    result = parse_package_config("example/valid/domain_one")
    assert result == PackageConfig(strict=True, tags=["one"])


def test_parse_valid_multi_tag_package_config():
    result = parse_package_config("example/valid/domain_two")
    assert result == PackageConfig(strict=False, tags=["two", "shared"])


def test_package_with_no_config():
    result = parse_package_config("example/")
    assert result is None


def test_invalid_project_config():
    with pytest.raises(ValidationError):
        parse_project_config("example/invalid/")


def test_empty_project_config():
    with pytest.raises(ValueError):
        parse_project_config("example/invalid/empty")


def test_invalid_package_config():
    with pytest.raises(ValidationError):
        parse_package_config("example/invalid")


def test_empty_package_config():
    with pytest.raises(ValueError):
        parse_package_config("example/invalid")



from __future__ import annotations


class TachError(Exception): ...


class TachParseError(TachError): ...


class TachSetupError(TachError): ...


__all__ = ["TachError", "TachParseError", "TachSetupError"]



from __future__ import annotations

import os
from dataclasses import dataclass, field
from typing import Optional

from tach import errors
from tach import filesystem as fs
from tach.colors import BCOLORS
from tach.constants import CONFIG_FILE_NAME, PACKAGE_FILE_NAME, TOOL_NAME
from tach.core import ProjectConfig
from tach.interactive import SelectedPackage, get_selected_packages_interactive
from tach.parsing import dump_project_config_to_yaml

__package_yml_template = """tags: ['{dir_name}']\n"""


@dataclass
class SetPackagesResult:
    package_paths: list[str] = field(default_factory=list)
    warnings: list[str] = field(default_factory=list)


def set_packages(
    selected_packages: list[SelectedPackage],
    path: str,
    exclude_paths: Optional[list[str]] = None,
) -> SetPackagesResult:
    package_paths: list[str] = []
    warnings: list[str] = []

    for existing_package, package_yml_path in fs.walk_configured_packages(
        root=path, exclude_paths=exclude_paths
    ):
        # If this package was not selected, we should delete the package.yml
        if not any(
            fs.canonical(selected_package.full_path) == fs.canonical(existing_package)
            for selected_package in selected_packages
        ):
            fs.delete_file(package_yml_path)

    for selected_package in selected_packages:
        init_py_path = os.path.join(selected_package.full_path, "__init__.py")
        if not os.path.exists(init_py_path):
            warnings.append(
                f"{BCOLORS.OKCYAN}Created __init__.py in selected package: '{selected_package.full_path}'{BCOLORS.ENDC}"
            )
            fs.write_file(init_py_path, f"# Generated by '{TOOL_NAME} init'")
        package_yml_path = os.path.join(
            selected_package.full_path, f"{PACKAGE_FILE_NAME}.yml"
        )
        package_paths.append(selected_package.full_path)
        if os.path.exists(package_yml_path):
            warnings.append(
                f"{BCOLORS.OKCYAN}Package file '{package_yml_path}' already exists.{BCOLORS.ENDC}"
            )
            continue
        package_yml_content = __package_yml_template.format(
            dir_name=fs.canonical(selected_package.full_path).replace(os.path.sep, ".")
        )
        fs.write_file(package_yml_path, package_yml_content)

    return SetPackagesResult(package_paths=package_paths, warnings=warnings)


@dataclass
class InitRootResult:
    warnings: list[str] = field(default_factory=list)


def init_root(root: str) -> InitRootResult:
    project_config_path = fs.get_project_config_path(root)
    if project_config_path:
        return InitRootResult(
            warnings=[
                f"{BCOLORS.OKCYAN}Project already contains {CONFIG_FILE_NAME}.yml{BCOLORS.ENDC}"
            ]
        )

    # Initialize an empty/default project configuration
    project_config = ProjectConfig()
    project_config_path = os.path.join(root, f"{CONFIG_FILE_NAME}.yml")
    config_yml_content = dump_project_config_to_yaml(project_config)
    fs.write_file(project_config_path, config_yml_content)

    return InitRootResult(warnings=[])


def pkg_edit_interactive(
    root: str, depth: Optional[int] = 1, exclude_paths: Optional[list[str]] = None
) -> tuple[bool, list[str]]:
    if not os.path.isdir(root):
        raise errors.TachSetupError(f"The path {root} is not a directory.")

    if exclude_paths is None:
        exclude_paths = ["tests/", "docs/"]

    warnings: list[str] = []

    # We only want to auto-select if the project appears not to have been configured already
    auto_select_initial_packages = fs.get_project_config_path(root) == ""
    selected_packages = get_selected_packages_interactive(
        root,
        depth=depth,
        exclude_paths=exclude_paths,
        auto_select_initial_packages=auto_select_initial_packages,
    )
    if selected_packages is not None:
        set_packages_result = set_packages(
            selected_packages=selected_packages, path=root, exclude_paths=exclude_paths
        )
        warnings.extend(set_packages_result.warnings)

        init_root_result = init_root(root)
        warnings.extend(init_root_result.warnings)
    else:
        return False, [f"{BCOLORS.OKCYAN}No changes saved.{BCOLORS.ENDC}"]

    return True, warnings
